# Development and Publishing

My aim is to use Test Driven Development (TDD) while developing this package.
Partly to learn TDD but also for what TDD aims to grant you (when done well):
a clear public API and trust that you won't unknowingly break it with new
features or refactorings.

## Environment Setup

Create a local virtualenv at the root of the repo:
```shell
python3.10 -m venv --upgrade-deps venv
```
This will create a virtualenv at `venv/`

Activate the virtualenv:
```shell
source venv/bin/activate
```
> **Note:** from now on all commands are assumed to be run
> within the activated virtualenv.

Install the project in editable mode and with all extras:
```shell
pip install -e '.[dev,doc,pre-commit,test]'
```

Install the [pre-commit hook](#pre-commit-hook):
```shell
pre-commit install
```

> **Note:** the pre-commit hook depends on the extra `test` also being installed
> in the venv. Unfortunately, you can't reference other extras under an extra
> when using just pyproject.toml's `[project.optional-dependencies]`, so this is
> a necessary compromise.

### pre-commit hook
[pre-commit][pre-commit] used as a git pre-commit hook. It runs several checks
and tests to verify that the commit adheres to the code style for consistency
and make sure tests are passing.

[pre-commit]: https://pre-commit.com/

Aside from running on git's pre-commit hook, it can be manually run on the
current working tree. It will run against staged files, only applying any
modifications if there are no unstaged ones, unless run with `--all-files`.
```shell
pre-commit run [hook]
```
> **Notice:** `[hook]` is optional. Substitute it to run a specific hook;
> without, all hooks are run.

> **Hint:** hooks are defined in
> [`.pre-commit-config.yaml`](.pre-commit-config.yaml).

If you want to run it on the dirty working tree, run it with `--all-files | -a`:
```shell
pre-commit run [hook] -a
```
> **Note:** pre-commit only ever runs on tracked files, even with `--all-files`.

## Local testing
pre-commit is configured to run the testing of `pytest`, `flake8` (for code
style and docstrings) and `mypy` (for static typechecking).

### pytest
Tests are written using [pytest][pytest] with the
[pytest-asyncio][pytest-asyncio] plugin to allow testing coroutines.

[pytest]: https://docs.pytest.org/en/latest/contents.html
[pytest-asyncio]: https://github.com/pytest-dev/pytest-asyncio

```shell
pytest
```

### tox
Run the `pytest` against the built package in isolated environments across
supported python versions, `mypy` and `flake8`.
```shell
tox
```

## Docs
Documentation are aimed to be generated from Google style docstrings in the
package using [sphinx][sphinx]. And also to later autogenerated and published
on [Read the Docs][rtd]. However, until we have an established public API
(`>=1.0.0`), this remains of less concern as the API is still being
established.

[sphinx]: https://www.sphinx-doc.org/en/master/
[rtd]: https://readthedocs.org/

### Building docs
Build the docs from Google style docstrings in the package.
```shell
tox -e docs
```

### Serving docs
Serve the built docs at <http://[::1]:8000> using python's
[http.server][http.server] module. Automatically attempt to open your web
browser at that URI using python's [webbrowser][webbrowser] module. Use
`Ctrl-C` to kill the webserver.
```shell
tox -e serve-docs
```

[http.server]: https://docs.python.org/3/library/http.server.html
[webbrowser]: https://docs.python.org/3/library/webbrowser.html

## Publishing

Building and publishing of the package to [PyPI][pypi] is handled
using [Flit][flit]. The project should adhere to [Semantic Versioning][semver]
and all that entails.

[pypi]: https://pypi.org/
[flit]: https://flit.pypa.io/en/latest/
[semver]: https://semver.org/spec/v2.0.0.html

### Code check
Flit does not allow publishing in a dirty working tree (i.e. not everything is
committed). If committed, pre-commit should've verified the code integrity for
the local version and _should_ be enough to not release anything broken.
However, testing all supported python versions by running `tox` is a good idea.
Finally, above tests are only run on the developers OS. Automated tests across
OS' are run on GitHub as a GitHub Action; for best peace of mind, make sure
those are successful as well.

### Update the Changelog
Update the [CHANGELOG](CHANGELOG.md) to detail changes in the new version.

### Update the Version
Update the package version in the code as given by [\_\_version\_\_][version].
This attribute is automatically used by Flit when building the package.

[version]: src/eventing/__init__.py#L9

### Create the commit
The `__version__` field in the package's `__init__.py` and CHANGELOG.md should
be the only two changes in a release commit.
```shell
git add CHANGELOG.md src/eventing/__init__.py
git commit -m 'Release <semver>'
```

### Build the package
Make sure the package builds using Flit:
```shell
flit build
```

### Add a Git Tag
Versions are tagged using annotated git tags:
```shell
git tag -a 'v<semver>' -m ''
```

Make sure to push the tag with:
```shell
git push --follow-tags
```

### Publish the package to PyPI
Publish the package to PyPI using Flit:
```shell
flit publish
```

Flit checks for credentials to publish looking for `~/.pypirc` with the `[pypi]`
label. See [Flit Upload docs][flit-upload-docs].

[flit-upload-docs]: https://flit.pypa.io/en/latest/upload.html
